# Surasith Boonaneksap File Created Aug 3rd 2021

from math import inf
from copy import deepcopy
import re

from PIL import Image, ImageDraw

from .turtle import Turtle
from .arg_handling import _arg_handling

class LSystem:
    """
    Store the state and evaluate an L-System

    Attributes
    ----------
    axiom : str
        The initial string or an axiom
    rules : dict {str : str}
        The production rules
    angle : float or int
        The angle ẟ (degrees) turned for each turning command
    dist : float or int
        The distance (pixels) travelled for each forward command
    iters : int
        The number of iterations
    result : str
        The result after rewriting
    result_img : Image
        The final image adfter drawing
    bound : tuple of tuple of ints
        The max and min of the image
    width : int
        The width of the drawing pen
    turtle : Turtle
        The current turtle state
    stack : list
        A stack used to support [ and ] command
    output : str
        The output filename
    """
    TESTING = False

    def __init__(self, axiom, rules, angle, dist, iters, width, output):
        """
        Initialize variables necessary for evaluating an L-system.
        Rewrite and draw a representation of that L-system.

        Parameters
        ----------
        axiom : str
            The initial string or an axiom
        rules : dict {str : str}
            The production rules
        angle : float or int
            The angle ẟ (degrees) turned for each turning command
        dist : float or int
            The distance (pixels) travelled for each forward command
        iters : int
            The number of iterations
        width : int
            The width of the drawing pen
        output : str
            The output file
        """
        
        _arg_handling(axiom, str)
        _arg_handling(rules, dict, 
                      [re.compile("."), re.compile(".*")],
                      err_msg="Invalid rules. "
                      "The predecessor can only have one symbol.")
        _arg_handling(angle, float)
        _arg_handling(dist, float)
        _arg_handling(iters, int, [0,inf], err_msg="The number of iterations "
                      "must be a positive integer.")
        _arg_handling(width, int, [0,inf], err_msg="The pen width must be a "
                      "positive integer.")
        if output:
            _arg_handling(output, str, re.compile(".+\..+"))

        self.axiom      = axiom
        self.rules      = rules
        self.angle      = angle
        self.dist       = dist
        self.iters      = iters

        self.result     = ""
        self.result_img = None
        self.bound      = ((0,0), (0,0)) # (max, min)
        self.width      = width

        self.turtle     = None
        
        self.stack      = []

        self.output     = output

        self._rewrite()
        self._find_bound()
        self._draw()
    
    def _rewrite(self):
        """
        Rewriting the axiom according to the given rules and arguments

        Symbols, including constants, without a production rule have an 
        identity property. That is the symbol will be converted to itself.

        Warnings
        --------
        This method is private and should only be called only once after
        initialization.
        
        Raise
        -----
        RuntimeError
            Raise when the resulting L-System have an incorrect number
        of brackets
        """

        self.result = self.axiom

        for i in range(self.iters):
            iter_result = ""

            for c in self.result:
                try:
                    iter_result += self.rules[c]
                except KeyError:
                    iter_result += c

            self.result = iter_result
        
        if(self.check_empty_stack(self.result)):
            raise RuntimeError("This L-system will pop an empty stack")
    
    def _eval(self, symbol):
        """
        Evaluate a symbol and command the turtle

        Parameter
        ---------
        symbol : char
            A variable or a constant

        Warnings
        --------
        This method modified the current state of the
        turtle and the stack.
        """

        if (symbol in "Ff"):
            self.turtle.forward(self.dist)
            
        elif (symbol == "+"):
            self.turtle.turn(-self.angle)

        elif (symbol == "-"):
            self.turtle.turn(+self.angle)

        elif (symbol == "["):
            self.stack.append( deepcopy(self.turtle) )
        
        elif (symbol == "]"):
            self.turtle = self.stack.pop()
        
        return self.turtle.pos

    def _find_bound(self):
        """
        Find the boundary of the canvas
        
        Warnings
        --------
        This method is private and should only be called
        after the rewriting process is completed.
        """
        
        self.turtle = Turtle()
        self.stack  = []

        max_ = [0, 0]
        min_ = [0, 0]

        for c in self.result:
            x, y = self._eval(c)

            max_ = [max(x, max_[0]), max(y, max_[1])]
            min_ = [min(x, min_[0]), min(y, min_[1])]
        
        self.bound = (tuple(max_), tuple(min_))

    def _draw(self):
        """
        Draw the lines as instructed on the string

        The output, if specified, will be saved to
        the given filename. Otherwise, the result
        will be shown after completion.
        
        Warnings
        --------
        This method is private and should only be called
        after the boundary of the image is found and the
        rewriting process is completed.
        
        Notes
        -----
        The image size after drawing must be at least (1, 1)
        """

        self.turtle = Turtle()
        self.stack = []

        img_width  = int(abs(self.bound[0][0] - self.bound[1][0]))
        img_height = int(abs(self.bound[0][1] - self.bound[1][1]))

        self.result_img = Image.new("1", (img_width, img_height))

        canvas = ImageDraw.Draw(self.result_img)

        _, min_ = self.bound
        
        start_x, start_y = 0, 0
        for c in self.result:
            x, y = self._eval(c)
            
            if c == "F":
                # Coordinates Correction
                # - The image's lowest point in the image is 0, 0
                # - y coordinate is inverted (for graphical coordinates)
                start_x_adj = start_x - min_[0]
                start_y_adj = img_height - (start_y - min_[1])

                x_adj       = x - min_[0]
                y_adj       = img_height - (y - min_[1])

                canvas.line([start_x_adj, start_y_adj, x_adj, y_adj], 
                            fill = 1, 
                            width = self.width)

            start_x, start_y = x, y
        
        if (not LSystem.TESTING):
            if (self.result_img.size[0] < 1 
                    or self.result_img.size[1] < 1):
                print("The image size must be at least (1,1)")
            else:
                print("done")
                if self.output:
                    self.result_img.save(self.output)
                else:
                    self.result_img.show()
              
    @staticmethod  
    def check_empty_stack(input):
        """
        Check if the L-system will pop an empty stack.
        
        As the L-system is being evaluated, there are more
        closing brackets ("[") than opening bracket ("]")
        at some point.
        
        Parameter
        ---------
        input: str
            An L-system string
        
        Return
        ------
        bool
            True if there's an attempt to pop an empty stack.
        False otherwise.
        """
        
        _arg_handling(input, str)
        
        unpaired_bracket = 0
        
        for c in input:
            if c == "[":
                unpaired_bracket += 1
            elif c == "]":
                unpaired_bracket -= 1
                
            if unpaired_bracket < 0:
                return True
        
        return False


# Surasith Boonaneksap File Created Aug 3rd 2021

from math import cos, sin, radians

class Turtle:
    """
    A class keeping a state of a turtle

    A turtle is an object used to represent a pen
    on a drawing board. It always move in a straight
    line according to its heading.

    Attributes
    ----------
    heading : float
        The current heading (degrees) of the turtle
    pos : tuple of ints
        The current coordinates, (x,y), of the turtle
    """

    def __init__(self, heading = 90, pos = (0,0)):
        """
        Initializing the turtle state

        Parameters
        ----------
        heading : float or int, default = 90
            The initial heading
        pos : tuple of ints, default = (0,0)
            The initial position
        """
        
        self.heading = heading % 360
        self.pos = pos # x, y
        
    def __eq__(self, obj):
        """Two Turtle objects are equal if all its attributes are equal"""
        
        return (self.heading == obj.heading and 
                self.pos[0] == obj.pos[0] and
                self.pos[1] == obj.pos[1])

    def forward(self, dist = 1):
        """
        Moving the turtle forward [dist] pixels in the heading direction
        
        Parameters
        ----------
        dist : float or int, default = 1
            The distance (pixels) that will be traveled
        """

        rad = radians(self.heading)
        self.pos = (self.pos[0] + cos(rad)*dist, 
                    self.pos[1] + sin(rad)*dist)
    
    def turn(self, deg = 90):
        """
        Change the heading of the turtle

        Parameters
        ----------
        deg : float or int, default: 90
            The angle (degrees) to change the current heading

        Notes
        -----
        The heading will always have a value between [0,360]
        """

        self.heading = (self.heading + deg) % 360


# Surasith Boonaneksap Aug 7th, 2021
from typing import Any
from re import Pattern, fullmatch

def _arg_handling(input: Any,
                  expected_type: Any = None, 
                  expected_value: Any = None, 
                  err_msg: str = None) -> None:
    """
    Check if a value/variable is of certain type and value.

    An exception is raised if the type or the value is not correct.
    This function only support an int, float, str, and dict{str: str}.
    The expected value can be exact, a range for numeric types using 
    a list (inclusive), or a regular expression for string 
    using re.Pattern.
    
    Parameters
    ----------
    input: Any
        A value or variable of the supported types
    expected_type: Any, default = None
        Expected input's type [int, float, str, dict]
    expected_value: Any, optional
        The expected value of input. If not specified, No value is checked.
    err_msg: str, optional
        The error message for an exception if an input is not correct.
        
    Raises
    ------
    ValueError
        Raise when the input doesn't have the expected value or the 
        parameters doesn't have the correct value.
    TypeError
        Raise when the input doesn't have the expected type or the 
        parameters doesn't have the correct type.

    Notes
    -----
    The key and value of dict must be string. A list of two regexes 
    can be used. The first one for the keys and the second for the values.

    Examples
    --------
    >>> import re
    >>> _arg_handling(5, int, 5)
    >>> _arg_handling([1,2,3], list, [1,2,3])
    Traceback (most recent call last):
        ...
    TypeError: list is not supported
    >>> _arg_handling(1, int, [0,6])
    >>> _arg_handling("email@email.com", str, re.compile(".+@.+"))
    >>> _arg_handling({"a":"A","b":"B"}, dict, \
    ... [re.compile("."), re.compile(".")])
    """

    supported_types = [int, float, str, dict]
    supported_expected_value_types = [*supported_types, list, Pattern]
    numeric = [int, float]
    
    #Exception: An int can be converted to float
    if (type(input) == int and expected_type == float):
        input = float(input)

    # The arguments' types must be supported
    if (type(input) not in supported_types):
        raise TypeError(f"{str(type(input))[8:-2]} is not supported")
    
    if (expected_type not in supported_types):
        raise TypeError("The provided expected type is not supported")
    
    if (type(err_msg) not in [str, type(None)]):
        print(type(err_msg))
        raise TypeError("The error message must be a string")
    
    # Check if the input's type matched the expected type
    if (type(input) != expected_type):
        raise TypeError(err_msg if err_msg else
                        f"Expected an input of type "
                        f"{str(expected_type)[8:-2]}")
    # Check the expected value if given
    elif expected_value:
        #Check the type of the expected value
        if (type(expected_value) not in supported_expected_value_types):
            raise TypeError(f"The type of the expected value is invalid")
        
        if (type(expected_value) in supported_types 
                and type(expected_value) != type(input)):
            raise TypeError("For an exact value comparison, the type of "
                            "the input and expected value must be the same")
        
        if (type(expected_value) == Pattern and type(input) != str):
            raise TypeError("A regular expression can only be used with "
                            "a string input")
            
        if (type(expected_value) == list 
                and type(input) not in [int, float, dict]):
            raise TypeError("A list can only be used with an int, a float, "
                            "or a dict")
        
        # Exact comparison
        elif (type(input) == type(expected_value)
                and input != expected_value):
            raise ValueError(err_msg if err_msg else
                                f"Expected an input's value: "
                                f"{expected_value}")
            
        # Regular Expression for string
        elif (type(input) == str and type(expected_value) == Pattern):
            if fullmatch(expected_value, input) == None:
                raise ValueError(err_msg if err_msg else
                                    f"No matching pattern found")
        
        # Range for int, float, and dict
        elif (type(expected_value) == list):
            if (len(expected_value) != 2):
                raise ValueError(f"The expected value must "
                                 f"have a list of size 2")
            else:
                if (type(input) == dict):
                    
                    for i in expected_value:
                        if type(i) != Pattern:
                            raise ValueError("The key and value must "
                                             "be pattern")
                            
                    for key in input:
                        if type(key) != str or type(input[key]) != str:
                            raise TypeError(err_msg if err_msg else \
                                "Keys and values must be string")
                        
                        if (fullmatch(expected_value[0], key) == None 
                                or fullmatch(expected_value[1], input[key])\
                                    == None):
                            raise ValueError(err_msg if err_msg else \
                                "Unexpected key or value")
                    
                elif (type(input) in [int, float]):
                    
                    for i in expected_value:
                        if type(i) not in [int, float]:
                            raise ValueError("The range of the expected "
                                             "values must be numeric")
                    
                    expected_value = sorted(expected_value)
                    
                    if (input < expected_value[0] 
                            or input > expected_value[1]):
                        raise ValueError(err_msg if err_msg else "Unexpected value")
                    
                else:
                    raise TypeError("list expected value is incompatible "
                                    "with a string input")


"""
Fractal generator using deterministic and context-free
L-system. (D0L)

This script is the entry point to command-line interface
of the software. Users must at least provide an axiom 
to an L-system in order to generate an image. Production 
rules, turning angle, distance, and the number of 
iterations are optional. 

Any unicode characters can be used as a variable except
the following constants that will be used to command
a turtle:

F = move forward and draw a line for d pixels
f = move forward without drawing a line for d pixels
+ = turn right by ẟ degrees
- = turn left by ẟ degrees
[ = push the turtle state to a stack
] = pop the turtle state out of the stack

This software requires Pillow 4.0.0 to draw and process
an image (May need a more recent version depending on
the Python's version)

Author: Surasith Boonaneksap

References:
- https://link-springer-com.ezproxy.lib.vt.edu/content/pdf/10.1007%2F978-1-4757-1428-9.pdf
- https://en.wikipedia.org/wiki/L-system
- https://jsantell.com/l-systems/
"""

#Surasith Boonaneksap Start on July 31st, 2021

__version__ = "1.0.0"

import argparse

from .lsystem import LSystem

DESCRIPTION = """

Fractal generator using deterministic and context-free
L-system. (D0L)

"""

def rules_parser(rules):
    """
    Convert a list of rules to a dictionary
    with predecessor as the key and successor
    as its value.

    Each rule must be in the form of
    "[predecessor]->[successor]"

    Parameter
    ---------
    rules: list of str
        A list containing production rules
    
    Return
    ------
    dict {str: str}
        The production rules in a dictionary
    """

    if not rules:
        return {}
    
    result = {}
    for r in rules:
        key, val = r.split("->")
        result[key] = val
    return result

# Fractal Generator with L-system
def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    
    parser.add_argument("-x", "--axiom", 
                        required=True, 
                        type=str, 
                        help="initial string or axiom, required")
    
    parser.add_argument("-r", "--rules", 
                        nargs='+', 
                        type=str, 
                        help='production rules. Uses "->" to '
                             'seperate predecessor from successor, optional')
    
    parser.add_argument("-a", "--angle", 
                        default=90, 
                        type=float, 
                        help="turning angle, ẟ (degrees), default = 90")
    
    parser.add_argument("-d", "--dist", 
                        default=100, 
                        type=float, 
                        help="distance (pixels), default = 100")
    
    parser.add_argument("-n", "--iters", 
                        default=0, 
                        type=int, 
                        help="number of iterations, default = 0")
    
    
    
    parser.add_argument("-o", "--output", 
                        type=str, 
                        help="output path. "
                             "If specified, the image will be saved "
                             "to the given path, optional")
    
    parser.add_argument("-w", "--width", 
                        default=1,
                        type=int, 
                        help="width of the drawing pen (pixels)"
                             ", default = 1")
    
    parser.add_argument("-v", "--version", 
                        action="version", 
                        version=f"lfract {__version__}")

    input = parser.parse_args()

    LSystem(input.axiom, 
            rules_parser(input.rules), 
            input.angle, 
            input.dist, 
            input.iters, 
            input.width, 
            input.output)

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=4.0']

entry_points = \
{'console_scripts': ['lfract = src.lfract:main']}

setup_kwargs = {
    'name': 'lfract',
    'version': '1.0.0rc1',
    'description': 'An open-source CLI software for generating fractal images using the L-System.',
    'long_description': '\n# L-System Fractal Generator\n\nAn open-source CLI software for generating fractal images using the L-System.\n\n## Features\n\n- Rewrite and draw a 2D deterministic context-free L-System. (D0L)\n- Support the following turtle commands:\n\n|  Command  | Description | \n| :-------- | :----------- |\n| F         | Draw a line forward *d* pixels in the current heading|\n| f         | Move forward *d* pixels in the current heading without drawing a line|\n| +         | Turn right by *δ* degrees |\n| -         | Turn left by *δ* degrees |\n| [         | Push the current turtle\'s state into a stack |\n| ]         | Pop a turtle\'s state out of the stack |\n\n- Support saving and previewing the output image\n- Support changing the drawing line\'s width\n\n## Background\n\n\nL-System (Lindenmayer System) is a string rewriting system that uses an iterative process. An L-System starts with ***an axiom***, an initial string, that will be rewritten following a set of ***production rules*** which contains pairs of ***predecessors*** and ***successors***. For each iteration, each symbol or character will be replaced with its corresponding successor. Any symbols without a production rule will be replaced by themselves. For example,\n\n> **Axiom:** N  \n> **Production Rules:**\n> |Predecessor|Successor|\n> |:---|:---|\n> |N|MNO|\n> |O|#|\n>\n> **Iteration 0:** N  \n> **Iteration 1:** MNO  \n> **Iteration 2:** MMNO#  \n> **Iteration 3:** MMMNO##  \n\nFor illustration, ```lfract``` reserves some characters for controlling a drawing pen called ***a turtle***. The turtle keeps track of its position and heading throughout the evaluation process. After some finite number of iterations, the resulting string will be evaluated and an image will be drawn. (Supported turtle commands are listed in the previous section)  \n\nThe current version of ```lfract``` supports only the deterministic and context-free version of the L-System. That is each rule has a probability of being applied of 1 and the location of each symbol in a string relative to other symbols doesn\'t matter.\n\n## Installation\n\nThis software is available to download on The Python Package Index (PyPI) and can be installed through ```pip```.  \n\nStart by making sure that ```pip``` is installed and up-to-date\n> ```python -m pip install --upgrade pip```\n\nRun the following command on your terminal\n\n> ```pip install lfract```\n\n## Usage\n\nInvokes the program using ```lfract```. There are seven optional arguments as follows:\n\n| Option |Description|\n| :--- | :----- |\n| -x, --axiom  | Initial string or axiom, required |\n| -r, --rules  | Production rules. Uses "->" to seperate predecessor from successor, optional |\n| -a, --angle  | Turning angle, *δ* (degrees), default = 90 |\n| -d, --dist   | Distance, *d* (pixels), default = 100 |\n| -n, --iters  | Number of iterations, default = 0 |\n| -o, --output | Output path. If specified, the image will be saved to the given path, optional |\n| -w, --width  | Width of the drawing pen (pixels), default = 1 |\n\nThe user must provide at least an axiom or the initial string to the L-System. All predecessors in the set of production rules must be unique. If there are duplicates, ```lfract``` will use the latest rule.\n\nFor the full usage description of ```lfract```, type:\n> ```lfract --help```\n\n## Examples\n\n1. Hilbert curve by Prusinkiewicz and Hanan [1]\n   > ```lfract -x X -r X->-YF+XFX+FY- Y->+XF-YFY-FX+ -a 90 -d 50 -n 5 -w 5```\n\n![hilbert](/images/hilbert.jpeg)\n\n2. Hexagonal tiling by Prusunkiewicz and Hanan [1]\n   > ```lfract -x X -r X->[-F+F[Y]+F][+F-F[X]-F] Y->[-F+F[Y]+F][+F-F-F] -a 60 -d 50 -n 10 -w 4```\n\n![hex](/images/hex.jpeg)\n\n3. A fractal by Surasith Boonaneksap\n   > ```lfract -x A -r A->fB+AF+Fff+Bf B->Bf-FffF-FfA-Afff -a 120 -n 7 -d 50 -w 5```\n\n![ex1](/images/ex1.jpeg)\n\n## Notes\n\n  - ```lfract``` adopts the L-System formal definition from [1]. It also follows the terminologies used in [1][2][3].\n  - Symbols in the L-System can be any Unicode characters.\n  - Currently not support:\n    - Stochastic L-System\n    - Context-sensitive L-System\n    - Parametric L-System\n\n## Dependency\n\nPillow 4.0.0 (Depending on the Python\'s version, a more recent version may be required. Check [Pillow\'s documentation](https://pillow.readthedocs.io/en/stable/installation.html#python-support))\n\n## License\n\nMIT\n\n## References\n\n[1]: Prunsinkiewicz, Przemyslaw and Hanan, James; "Lecture Notes in Biomathematics: Lindenmayer Systems, Fractals, and Plants" [link](http://algorithmicbotany.org/papers/lsfp.pdf)  \n[2]: Wikipedia; "L-system" [link](https://en.wikipedia.org/wiki/L-system)  \n[3]: Santel, Jordan; "L-systems" [link](https://jsantell.com/l-systems/)  ',
    'author': 'Surasith Boonaneksap',
    'author_email': 'surasith.boo@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SurasithO/lfract',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
